# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
[
  {
    "test_case_name": "first",
    "arguments": {
      "${username}": "demo",
      "${password}": "mode"
    },
    "tags": [
      "tag1",
      "tag2",
      "smoke"
    ],
    "documentation": "This is the doc"
  },
  {
    "test_case_name": "second",
    "arguments": {
      "${username}": "${EMPTY}",
      "${password}": "mode"
    },
    "tags": [
      "tag1",
      "smoke"
    ],
    "documentation": "This is the doc"
  }
]
"""

from .AbstractReaderClass import AbstractReaderClass


class json_reader(AbstractReaderClass):
    pass



"""Pre-run modifier that excludes tests that run PASS last time.
"""

import os
import re

from robot.api import ExecutionResult, ResultVisitor, SuiteVisitor  # type: ignore
from robot.running.model import Variable  # type: ignore


class rerunfailed(SuiteVisitor):
    def __init__(self, original_output_xml):
        if not os.path.isfile(original_output_xml):
            raise FileNotFoundError(f"{original_output_xml} is no file")
        result = ExecutionResult(original_output_xml)
        results_visitor = DataDriverResultsVisitor()
        result.visit(results_visitor)
        self._failed_tests = results_visitor.failed_tests

    def start_suite(self, suite):
        """Remove tests that match the given pattern."""
        if self.has_no_tests(suite.name):
            suite.tests.clear()
            return None
        if self._suite_is_data_driven(suite):
            dynamic_tests = Variable("@{DYNAMICTESTS}", self._failed_tests, suite.source)
            suite.resource.variables.append(dynamic_tests)
        else:
            suite.tests = [
                t for t in suite.tests if f"{t.parent.name}.{t.name}" in self._failed_tests
            ]

    def has_no_tests(self, name):
        r = re.compile(f'^{re.escape(f"{name}.")}.*$')
        return not bool(list(filter(r.match, self._failed_tests)))

    def _suite_is_data_driven(self, suite):
        for resource in suite.resource.imports:
            if resource.name == "DataDriver":
                return True

    def end_suite(self, suite):
        """Remove suites that are empty after removing tests."""
        suite.suites = [s for s in suite.suites if s.test_count > 0]

    def visit_test(self, test):
        """Avoid visiting tests and their keywords to save a little time."""
        pass


class DataDriverResultsVisitor(ResultVisitor):
    def __init__(self):
        self.failed_tests = list()

    def start_test(self, test):
        if test.status == "FAIL":
            self.failed_tests.append(f"{test.parent.name}.{test.name}")



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import csv
from DataDriver.AbstractReaderClass import AbstractReaderClass


class csv_reader(AbstractReaderClass):
    def get_data_from_source(self):
        self._register_dialects()
        self._read_file_to_data_table()
        return self.data_table

    def _register_dialects(self):
        if self.csv_dialect.lower() == "userdefined":
            csv.register_dialect(
                self.csv_dialect,
                delimiter=self.delimiter,
                quotechar=self.quotechar,
                escapechar=self.escapechar,
                doublequote=self.doublequote,
                skipinitialspace=self.skipinitialspace,
                lineterminator=self.lineterminator,
                quoting=csv.QUOTE_ALL,
            )
        elif self.csv_dialect == "Excel-EU":
            csv.register_dialect(
                self.csv_dialect,
                delimiter=";",
                quotechar='"',
                escapechar="\\",
                doublequote=True,
                skipinitialspace=False,
                lineterminator="\r\n",
                quoting=csv.QUOTE_ALL,
            )

    def _read_file_to_data_table(self):
        with open(self.file, "r", encoding=self.csv_encoding) as csvfile:
            reader = csv.reader(csvfile, self.csv_dialect)
            for row_index, row in enumerate(reader):
                try:
                    if row_index == 0:
                        self._analyse_header(row)
                    else:
                        self._read_data_from_table(row)
                except Exception as e:
                    e.row = row_index + 1
                    raise e



from DataDriver import DataDriver
from inspect import getdoc

with open("Readme.rst", "w", encoding="utf-8") as readme:
    doc_string = getdoc(DataDriver)
    readme.write(str(doc_string).replace("\\", "\\\\").replace("\\\\*", "\\*"))



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    import pandas as pd  # type: ignore
    from math import nan  # type: ignore
    import xlrd  # type: ignore
except ImportError:
    raise ImportError(
        """Requirements (pandas, xlrd) for XLS support are not installed.
    Use 'pip install -U robotframework-datadriver[XLS]' to install XLS support."""
    )

from .xlsx_reader import xlsx_reader


class xls_reader(xlsx_reader):
    def read_data_frame_from_file(self, dtype):
        return pd.read_excel(self.file, sheet_name=self.sheet_name, dtype=dtype).replace(
            nan, "", regex=True
        )



# Copyright 2021-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This Reader has initialy been created by Samual Montgomery-Blinn (https://github.com/montsamu)
# Thanks for this contribution


from os.path import normpath, basename, splitext, abspath, isfile, isdir
from glob import glob

from DataDriver.AbstractReaderClass import AbstractReaderClass


class glob_reader(AbstractReaderClass):
    def get_data_from_source(self):
        self._read_glob_to_data_table()
        return self.data_table

    def _read_glob_to_data_table(self):
        self._analyse_header(["*** Test Cases ***", self.kwargs.get("arg_name", "${file_name}")])
        for match_path in sorted(glob(self.file)):
            match_path = abspath(match_path).replace("\\", "/")
            if isfile(match_path):
                test_case_name = basename(splitext(match_path)[0])
            elif isdir(match_path):
                test_case_name = basename(normpath(match_path))
            else:
                test_case_name = match_path
            self._read_data_from_table([test_case_name, match_path])



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Any, Dict, List

from robot.utils import DotDict  # type: ignore

from .utils import PabotOpt, TagHandling


class ReaderConfig:

    TEST_CASE_TABLE_NAME = "*** Test Cases ***"

    def __init__(
        self,
        file: Optional[str] = None,
        encoding: Optional[str] = None,
        dialect: Optional[str] = None,
        delimiter: Optional[str] = None,
        quotechar: Optional[str] = None,
        escapechar: Optional[str] = None,
        doublequote: Optional[bool] = None,
        skipinitialspace: Optional[bool] = None,
        lineterminator: Optional[str] = None,
        sheet_name: Any = None,
        reader_class: Optional[str] = None,
        file_search_strategy: str = "path",
        file_regex: Optional[str] = None,
        include: Optional[str] = None,
        exclude: Optional[str] = None,
        handle_template_tags: TagHandling = TagHandling.UnsetTags,
        list_separator: Optional[str] = ",",
        config_keyword: Optional[str] = None,
        optimize_pabot: PabotOpt = PabotOpt.Equal,
        **kwargs
    ):

        self.file = file
        self.encoding = encoding
        self.dialect = dialect
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.doublequote = doublequote
        self.skipinitialspace = skipinitialspace
        self.lineterminator = lineterminator
        self.sheet_name = sheet_name
        self.reader_class = reader_class
        self.file_search_strategy = file_search_strategy
        self.file_regex = file_regex
        self.include = include
        self.exclude = exclude
        self.handle_template_tags = handle_template_tags
        self.list_separator = list_separator
        self.config_keyword = config_keyword
        self.optimize_pabot = optimize_pabot
        self.kwargs = kwargs


class TestCaseData(DotDict):
    def __init__(
        self,
        test_case_name: str = "",
        arguments: Optional[Dict] = None,
        tags: Optional[List] = None,
        documentation: str = "",
    ):
        super().__init__()
        self.test_case_name = test_case_name
        self.arguments = arguments if arguments else {}
        self.tags = tags
        self.documentation = documentation



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import csv
import time
import os
from .AbstractReaderClass import AbstractReaderClass, ReaderConfig


class pict_reader(AbstractReaderClass):
    def __init__(self, reader_config: ReaderConfig):
        super().__init__(reader_config)
        file = self.file or ""
        self.pictout_file = f"{os.path.splitext(file)[0]}{time.time()}.pictout"

    def get_data_from_source(self):
        self._register_dialect()
        self._create_gemerated_file_from_model_file()
        self._read_generated_file_to_dictionaries()
        return self.data_table

    @staticmethod
    def _register_dialect():
        csv.register_dialect(
            "PICT",
            delimiter="\t",
            skipinitialspace=False,
            lineterminator="\r\n",
            quoting=csv.QUOTE_NONE,
        )

    def _create_gemerated_file_from_model_file(self):
        os.system(f'pict "{self.file}" > "{self.pictout_file}"')

    def _read_generated_file_to_dictionaries(self):
        with open(self.pictout_file, "r", encoding="utf_8") as csvfile:
            reader = csv.reader(csvfile, "PICT")
            for row_index, row in enumerate(reader):
                if row_index == 0:
                    row_of_variables = []
                    for cell in row:
                        row_of_variables.append(f"${{{cell}}}")
                    self._analyse_header(row_of_variables)
                else:
                    self._read_data_from_table(row)
        os.remove(self.pictout_file)



import sys

from enum import IntEnum

from robot.run import USAGE  # type: ignore
from robot.utils.argumentparser import ArgumentParser  # type: ignore

SINGLE_ARG_CHARACTERS = ".?hTX"


class ArgumentState(IntEnum):
    ANALYZE_NEXT = 0
    ONE_KNOWN = 1
    TWO_KNOWN = 2


def robot_options():
    arg_parser = ArgumentParser(
        USAGE, auto_pythonpath=False, auto_argumentfile=True, env_options="ROBOT_OPTIONS"
    )
    valid_args = filter_args(arg_parser)
    return arg_parser.parse_args(valid_args)[0]


def filter_args(arg_parser):
    short_opts = arg_parser._short_opts
    long_opts = arg_parser._long_opts
    arg_state = ArgumentState.ANALYZE_NEXT
    valid_robot_args = list()
    for arg in sys.argv[1:]:
        if arg_state == ArgumentState.ANALYZE_NEXT:
            arg_state = get_argument_state(arg, short_opts, long_opts)
        if arg_state >= ArgumentState.ONE_KNOWN:
            valid_robot_args.append(arg)
            arg_state -= 1
    return valid_robot_args


def get_argument_state(arg, short_opts, long_opts):
    param_opt = [l_opt[:-1] for l_opt in long_opts if l_opt[-1:] == "="]
    arg_state = 0
    if is_short_option(arg):
        if arg[1] in SINGLE_ARG_CHARACTERS:
            arg_state = ArgumentState.ONE_KNOWN
        elif arg[1] in short_opts:
            arg_state = ArgumentState.TWO_KNOWN
    elif is_long_option(arg):
        if arg[2:] in param_opt:
            arg_state = ArgumentState.TWO_KNOWN
        elif arg[2:] in long_opts:
            arg_state = ArgumentState.ONE_KNOWN
    return arg_state


def is_short_option(arg):
    return len(arg) == 2 and arg[0] == "-"


def is_long_option(arg):
    return len(arg) > 2 and arg[:2] == "--"


def is_pabot_testlevelsplit():
    return "--testlevelsplit" in sys.argv



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from re import compile
from typing import List

from .ReaderConfig import ReaderConfig
from .ReaderConfig import TestCaseData
from .search import search_variable

from robot.libraries.BuiltIn import BuiltIn  # type: ignore
from robot.utils import DotDict  # type: ignore


built_in = BuiltIn()


class AbstractReaderClass(ABC):
    def __init__(self, reader_config: ReaderConfig):

        self.reader_config = reader_config
        self.file = reader_config.file
        self.csv_encoding = reader_config.encoding
        self.csv_dialect = reader_config.dialect
        self.delimiter = reader_config.delimiter
        self.quotechar = reader_config.quotechar
        self.escapechar = reader_config.escapechar
        self.doublequote = reader_config.doublequote
        self.skipinitialspace = reader_config.skipinitialspace
        self.lineterminator = reader_config.lineterminator
        self.sheet_name = reader_config.sheet_name
        self.list_separator = reader_config.list_separator
        self.handle_template_tags = reader_config.handle_template_tags
        self.kwargs = reader_config.kwargs
        for key, value in reader_config.kwargs.items():
            setattr(self, key, value)

        self.test_case_column_id = None
        self.arguments_column_ids: List = list()
        self.tags_column_id = None
        self.documentation_column_id = None
        self.header: List = list()
        self.data_table: List = list()

        self.TESTCASE_TABLE_NAME = ReaderConfig.TEST_CASE_TABLE_NAME
        self.TEST_CASE_TABLE_PATTERN = compile(r"(?i)^(\*+\s*test ?cases?[\s*].*)")
        self.TASK_TABLE_PATTERN = compile(r"(?i)^(\*+\s*tasks?[\s*].*)")
        self.VARIABLE_PATTERN = compile(r"([$@&e]\{)(.*?)(\})")
        self.TAGS_PATTERN = compile(r"(?i)(\[)(tags)(\])")
        self.DOCUMENTATION_PATTERN = compile(r"(?i)(\[)(documentation)(\])")
        self.LIT_EVAL_PATTERN = compile(r"e\{(.+)\}")

    def get_data_from_source(self):
        raise NotImplementedError(
            "This method should be implemented and return self.data_table to DataDriver..."
        )

    def _is_test_case_header(self, header_string: str):
        is_test = self.TEST_CASE_TABLE_PATTERN.fullmatch(header_string)
        is_task = self.TASK_TABLE_PATTERN.fullmatch(header_string)
        return is_task or is_test

    def _is_variable(self, header_string: str):
        return self.VARIABLE_PATTERN.match(header_string)

    def _is_tags(self, header_string: str):
        return self.TAGS_PATTERN.match(header_string)

    def _is_documentation(self, header_string: str):
        return self.DOCUMENTATION_PATTERN.match(header_string)

    def _analyse_header(self, header_cells):
        self.header = header_cells
        for cell_index, cell in enumerate(self.header):
            if self._is_test_case_header(cell):
                self.test_case_column_id = cell_index
            elif self._is_variable(cell):
                self.arguments_column_ids.append(cell_index)
            elif self._is_tags(cell):
                self.tags_column_id = cell_index
            elif self._is_documentation(cell):
                self.documentation_column_id = cell_index

    def _read_data_from_table(self, row):
        test_case_name = (
            row[self.test_case_column_id] if self.test_case_column_id is not None else ""
        )
        arguments = {}
        for arguments_column_id in self.arguments_column_ids:
            variable_string = self.header[arguments_column_id]
            variable_value = row[arguments_column_id]
            if self.LIT_EVAL_PATTERN.fullmatch(variable_string):
                variable_string = f"${variable_string[1:]}"
                variable_value = built_in.replace_variables(variable_value)
                variable_value = built_in.evaluate(variable_value)
            variable_match = search_variable(variable_string)
            if variable_match.is_variable:
                arguments.update(
                    self._get_arguments_entry(variable_match, variable_value, arguments)
                )
        tags = (
            [t.strip() for t in row[self.tags_column_id].split(",")]
            if self.tags_column_id
            else None
        )
        documentation = row[self.documentation_column_id] if self.documentation_column_id else ""

        self.data_table.append(TestCaseData(test_case_name, arguments, tags, documentation))

    def _get_arguments_entry(self, variable_match, variable_value, arguments):
        base = variable_match.base
        items = variable_match.items
        if variable_match.is_list_variable:
            if not variable_value:
                variable_value = []
            else:
                variable_value = [
                    built_in.replace_variables(var)
                    for var in (str(variable_value).split(self.list_separator))
                ]
        elif variable_match.is_dict_variable:
            variable_value = built_in.create_dictionary(
                *(str(variable_value).split(self.list_separator))
            )
        if "." in base:  # is dot notated advanced variable dictionary ${dict.key.subkey}
            base, *items = base.split(".")
            variable_value = self._update_argument_dict(arguments, base, items, variable_value)
        elif items:  # is dictionary syntax ${dict}[key][subkey]
            variable_value = self._update_argument_dict(arguments, base, items, variable_value)

        return {self._as_var(base): variable_value}

    def _update_argument_dict(self, arguments, base, items, value):
        if self._as_var(base) not in arguments:
            arguments[self._as_var(base)] = built_in.create_dictionary()
        argument = arguments[self._as_var(base)]

        if isinstance(argument, DotDict):
            selected_key = argument
            for key in items:
                if key != items[-1]:
                    if key not in selected_key or not isinstance(selected_key[key], DotDict):
                        selected_key[key] = built_in.create_dictionary()
                    selected_key = selected_key[key]
            selected_key[items[-1]] = built_in.replace_variables(value)
            return argument
        else:
            raise TypeError(f"{self._as_var(base)} is defined with a wrong type. Not defaultdict.")

    def _as_var(self, base):
        return f"${{{base}}}"



import math

from enum import Enum, auto
from robot.libraries.BuiltIn import BuiltIn  # type: ignore
from robot.api import logger  # type: ignore
from typing import List, Any

from .argument_utils import is_pabot_testlevelsplit


class Encodings(Enum):
    """
    Python comes with a number of codecs built-in,
    either implemented as C functions or with dictionaries as mapping tables.
    The following table lists the codecs by name,
    together with a few common aliases, and the languages for which the encoding is likely used.
    Neither the list of aliases nor the list of languages is meant to be exhaustive.
    Notice that spelling alternatives that only differ in case or use a hyphen instead
    of an underscore are also valid aliases; therefore, e.g. ``utf-8` is a valid alias for the ``utf_8`` codec.

    *CPython implementation detail:* Some common encodings can bypass the codecs lookup machinery to improve performance.
    These optimization opportunities are only recognized by CPython for a limited set of (case insensitive) aliases:
    utf-8, utf8, latin-1, latin1, iso-8859-1, iso8859-1, mbcs (Windows only),
    ascii, us-ascii, utf-16, utf16, utf-32, utf32, and the same using underscores instead of dashes.
    Using alternative aliases for these encodings may result in slower execution.

    Changed in version 3.6: Optimization opportunity recognized for us-ascii.

    Many of the character sets support the same languages. They vary in individual characters (e.g. whether the EURO SIGN is supported or not), and in the assignment of characters to code positions. For the European languages in particular, the following variants typically exist:

    - utf-8
    - cp1252
    - an ISO 8859 codeset
    - a Microsoft Windows code page, which is typically derived from an 8859 codeset, but replaces control characters with additional graphic characters
    - an IBM EBCDIC code page
    - an IBM PC code page, which is ASCII compatible
    """

    ascii = auto()
    big5 = auto()
    big5hkscs = auto()
    cp037 = auto()
    cp273 = auto()
    cp424 = auto()
    cp437 = auto()
    cp500 = auto()
    cp720 = auto()
    cp737 = auto()
    cp775 = auto()
    cp850 = auto()
    cp852 = auto()
    cp855 = auto()
    cp856 = auto()
    cp857 = auto()
    cp858 = auto()
    cp860 = auto()
    cp861 = auto()
    cp862 = auto()
    cp863 = auto()
    cp864 = auto()
    cp865 = auto()
    cp866 = auto()
    cp869 = auto()
    cp874 = auto()
    cp875 = auto()
    cp932 = auto()
    cp949 = auto()
    cp950 = auto()
    cp1006 = auto()
    cp1026 = auto()
    cp1125 = auto()
    cp1140 = auto()
    cp1250 = auto()
    cp1251 = auto()
    cp1252 = auto()
    cp1253 = auto()
    cp1254 = auto()
    cp1255 = auto()
    cp1256 = auto()
    cp1257 = auto()
    cp1258 = auto()
    euc_jp = auto()
    euc_jis_2004 = auto()
    euc_jisx0213 = auto()
    euc_kr = auto()
    gb2312 = auto()
    gbk = auto()
    gb18030 = auto()
    hz = auto()
    iso2022_jp = auto()
    iso2022_jp_1 = auto()
    iso2022_jp_2 = auto()
    iso2022_jp_2004 = auto()
    iso2022_jp_3 = auto()
    iso2022_jp_ext = auto()
    iso2022_kr = auto()
    latin_1 = auto()
    iso8859_2 = auto()
    iso8859_3 = auto()
    iso8859_4 = auto()
    iso8859_5 = auto()
    iso8859_6 = auto()
    iso8859_7 = auto()
    iso8859_8 = auto()
    iso8859_9 = auto()
    iso8859_10 = auto()
    iso8859_11 = auto()
    iso8859_13 = auto()
    iso8859_14 = auto()
    iso8859_15 = auto()
    iso8859_16 = auto()
    johab = auto()
    koi8_r = auto()
    koi8_t = auto()
    koi8_u = auto()
    kz1048 = auto()
    mac_cyrillic = auto()
    mac_greek = auto()
    mac_iceland = auto()
    mac_latin2 = auto()
    mac_roman = auto()
    mac_turkish = auto()
    ptcp154 = auto()
    shift_jis = auto()
    shift_jis_2004 = auto()
    shift_jisx0213 = auto()
    utf_32 = auto()
    utf_32_be = auto()
    utf_32_le = auto()
    utf_16 = auto()
    utf_16_be = auto()
    utf_16_le = auto()
    utf_7 = auto()
    utf_8 = auto()
    utf_8_sig = auto()


class PabotOpt(Enum):
    """
    You can switch Pabot --testlevelsplit between three modes:

    - Equal: means it creates equal sizes groups
    - Binary: is more complex. it created a decreasing size of containers to support better balancing.
    - Atomic: it does not group tests at all and runs really each test case in a separate thread.

    See `Pabot and DataDriver <#pabot-and-datadriver>`__ for more details.

    This can be set by ``optimize_pabot`` in Library import.
    """

    Equal = auto()
    Binary = auto()
    Atomic = auto()


class TagHandling(Enum):
    """
    You can configure how to handle tags from the template in generated tests:

    - ForceTags: Will add all tags of the template test to all data-driven test cases.
    - UnsetTags: Will add only these tags from template test to the tests, that are not assigned to any of the test cases.
    - DefaultTags: Will only add tags to data-driven test if no tag is set on that test.
    - NoTags: Will not add any tags from the template to date-driven tests.

    """

    ForceTags = auto()
    UnsetTags = auto()
    DefaultTags = auto()
    NoTags = auto()


def debug(msg: Any, newline: bool = True, stream: str = "stdout"):
    if get_variable_value("${LOG LEVEL}") in ["DEBUG", "TRACE"]:
        logger.console(msg, newline, stream)


def console(msg: Any, newline: bool = True, stream: str = "stdout"):
    logger.console(msg, newline, stream)


def warn(msg: Any, html: bool = False):
    logger.warn(msg, html)


def error(msg: Any, html: bool = False):
    logger.error(msg, html)


def get_filter_dynamic_test_names():
    dynamic_test_list = get_variable_value("${DYNAMICTESTS}")
    if isinstance(dynamic_test_list, str):
        return dynamic_test_list.split("|")
    elif isinstance(dynamic_test_list, list):
        return dynamic_test_list
    else:
        dynamic_test_name = get_variable_value("${DYNAMICTEST}")
        if dynamic_test_name:
            BuiltIn().set_suite_metadata("DataDriver", dynamic_test_name, True)
            return [dynamic_test_name]


def is_pabot_dry_run():
    return is_pabot_testlevelsplit() and get_variable_value("${PABOTQUEUEINDEX}") == "-1"


def get_variable_value(name: str):
    return BuiltIn().get_variable_value(name)


def is_same_keyword(first: str, second: str):
    return _get_normalized_keyword(first) == _get_normalized_keyword(second)


def _get_normalized_keyword(keyword: str):
    return keyword.lower().replace(" ", "").replace("_", "")


def binary_partition_test_list(test_list: List, process_count: int):
    fractions = equally_partition_test_list(test_list, process_count)
    return_list = list()
    for i in range(int(math.sqrt(len(test_list) // process_count))):
        first, second = _partition_second_half(fractions)
        return_list.extend(first)
        fractions = second
    return_list.extend(fractions)
    return [test_name for test_name in return_list if test_name]


def _partition_second_half(fractions):
    first = fractions[: len(fractions) // 2]
    second = list()
    for sub_list in fractions[len(fractions) // 2 :]:
        sub_sub_list = equally_partition_test_list(sub_list, 2)
        second.extend(sub_sub_list)
    return first, second


def equally_partition_test_list(test_list: List, fraction_count: int):
    quotient, remainder = divmod(len(test_list), fraction_count)
    return [
        test_list[i * quotient + min(i, remainder) : (i + 1) * quotient + min(i + 1, remainder)]
        for i in range(fraction_count)
    ]



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

from robot.errors import VariableError  # type: ignore
from robot.utils import py2to3  # type: ignore


def search_variable(string, identifiers="$@&%*", ignore_errors=False):
    if not (isinstance(string, str) and "{" in string):
        return VariableMatch(string)
    return VariableSearcher(identifiers, ignore_errors).search(string)


@py2to3
class VariableMatch(object):
    def __init__(self, string, identifier=None, base=None, items=(), start=-1, end=-1):
        self.string = string
        self.identifier = identifier
        self.base = base
        self.items = items
        self.start = start
        self.end = end

    def resolve_base(self, variables, ignore_errors=False):
        if self.identifier:
            internal = search_variable(self.base)
            self.base = variables.replace_string(
                internal,
                custom_unescaper=unescape_variable_syntax,
                ignore_errors=ignore_errors,
            )

    @property
    def name(self):
        return "%s{%s}" % (self.identifier, self.base) if self else None

    @property
    def before(self):
        return self.string[: self.start] if self.identifier else self.string

    @property
    def match(self):
        return self.string[self.start : self.end] if self.identifier else None

    @property
    def after(self):
        return self.string[self.end :] if self.identifier else None

    @property
    def is_variable(self):
        return bool(
            self.identifier and self.base and self.start == 0 and self.end == len(self.string)
        )

    @property
    def is_list_variable(self):
        return bool(self.is_variable and self.identifier == "@" and not self.items)

    @property
    def is_dict_variable(self):
        return bool(self.is_variable and self.identifier == "&" and not self.items)

    def __nonzero__(self):
        return self.identifier is not None

    def __unicode__(self):
        if not self:
            return "<no match>"
        items = "".join("[%s]" % i for i in self.item) if self.items else ""
        return "%s{%s}%s" % (self.identifier, self.base, items)


class VariableSearcher(object):
    def __init__(self, identifiers, ignore_errors=False):
        self.identifiers = identifiers
        self._ignore_errors = ignore_errors
        self.start = -1
        self.variable_chars = []
        self.item_chars = []
        self.items = []
        self._open_brackets = 0  # Used both with curly and square brackets
        self._escaped = False

    def search(self, string):
        if not self._search(string):
            return VariableMatch(string)
        match = VariableMatch(
            string=string,
            identifier=self.variable_chars[0],
            base="".join(self.variable_chars[2:-1]),
            start=self.start,
            end=self.start + len(self.variable_chars),
        )
        if self.items:
            match.items = tuple(self.items)
            match.end += sum(len(i) for i in self.items) + 2 * len(self.items)
        return match

    def _search(self, string, offset=0):
        start = self._find_variable_start(string)
        if start == -1:
            return False
        self.start = start + offset
        self._open_brackets += 1
        self.variable_chars = [string[start], "{"]
        start += 2
        state = self.variable_state
        for char in string[start:]:
            state = state(char)
            self._escaped = False if char != "\\" else not self._escaped
            if state is None:
                break
        if state:
            try:
                self._validate_end_state(state)
            except VariableError:
                if self._ignore_errors:
                    return False
                raise
        return True

    def _find_variable_start(self, string):
        start = 1
        while True:
            start = string.find("{", start) - 1
            if start < 0:
                return -1
            if self._start_index_is_ok(string, start):
                return start
            start += 2

    def _start_index_is_ok(self, string, index):
        return string[index] in self.identifiers and not self._is_escaped(string, index)

    def _is_escaped(self, string, index):
        escaped = False
        while index > 0 and string[index - 1] == "\\":
            index -= 1
            escaped = not escaped
        return escaped

    def variable_state(self, char):
        self.variable_chars.append(char)
        if char == "}" and not self._escaped:
            self._open_brackets -= 1
            if self._open_brackets == 0:
                if not self._can_have_items():
                    return None
                return self.waiting_item_state
        elif char == "{" and not self._escaped:
            self._open_brackets += 1
        return self.variable_state

    def _can_have_items(self):
        return self.variable_chars[0] in "$@&"

    def waiting_item_state(self, char):
        if char == "[":
            self._open_brackets += 1
            return self.item_state
        return None

    def item_state(self, char):
        if char == "]" and not self._escaped:
            self._open_brackets -= 1
            if self._open_brackets == 0:
                self.items.append("".join(self.item_chars))
                self.item_chars = []
                # Don't support chained item access with old @ and & syntax.
                # The old syntax was deprecated in RF 3.2 and in RF 3.3 it'll
                # be reassigned to mean using item in list/dict context.
                if self.variable_chars[0] in "@&":
                    return None
                return self.waiting_item_state
        elif char == "[" and not self._escaped:
            self._open_brackets += 1
        self.item_chars.append(char)
        return self.item_state

    def _validate_end_state(self, state):
        if state == self.variable_state:
            incomplete = "".join(self.variable_chars)
            raise VariableError("Variable '%s' was not closed properly." % incomplete)
        if state == self.item_state:
            variable = "".join(self.variable_chars)
            items = "".join("[%s]" % i for i in self.items)
            incomplete = "".join(self.item_chars)
            raise VariableError(
                "Variable item '%s%s[%s' was not closed "
                "properly." % (variable, items, incomplete)
            )


def unescape_variable_syntax(item):
    def handle_escapes(match):
        escapes, text = match.groups()
        if len(escapes) % 2 == 1 and starts_with_variable_or_curly(text):
            return escapes[1:]
        return escapes

    def starts_with_variable_or_curly(text):
        if text[0] in "{}":
            return True
        match = search_variable(text, ignore_errors=True)
        return match and match.start == 0

    return re.sub(r"(\\+)(?=(.+))", handle_escapes, item)


# TODO: This is pretty odd/ugly and used only in two places. Implement
# something better or just remove altogether.
@py2to3
class VariableIterator(object):
    def __init__(self, string, identifiers="$@&%*"):
        self._string = string
        self._identifiers = identifiers

    def __iter__(self):
        remaining = self._string
        while True:
            match = search_variable(remaining, self._identifiers)
            if not match:
                break
            remaining = match.after
            yield match.before, match.match, remaining

    def __len__(self):
        return sum(1 for _ in self)

    def __nonzero__(self):
        try:
            next(iter(self))
        except StopIteration:
            return False
        else:
            return True



from setuptools import setup
from setuptools import find_packages
import re
from os.path import abspath, dirname, join

CURDIR = dirname(abspath(__file__))

with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(join(CURDIR, "src", "DataDriver", "DataDriver.py"), encoding="utf-8") as f:
    VERSION = re.search('\n__version__ = "(.*)"', f.read()).group(1)

setup(
    name="robotframework-datadriver",
    version=VERSION,
    author="René Rohner(Snooz82)",
    author_email="snooz@posteo.de",
    description="A library for Data-Driven Testing.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/Snooz82/robotframework-datadriver",
    package_dir={"": "src"},
    packages=find_packages("src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Testing :: Acceptance",
        "Framework :: Robot Framework",
    ],
    install_requires=["robotframework >= 3.1", "docutils", "Pygments"],
    extras_require={"xls": ["pandas", "xlrd >= 1.2.0", "openpyxl"]},
    python_requires=">=3.6.1",
)



#!/usr/bin/env python

"""Simple HTTP server for Robot Framework web testing demo.

Usage:  server.py [port]

This server serves HTML pages under `html` directory. Server is started simply
by running this script from the command line or double-clicking it in a file
manager. In the former case the server can be shut down with Ctrl-C and in the
latter case by closing the opened window.

By default the server uses port 7272, but a custom port can be given as
an argument from the command line.
"""

from os import chdir
from os.path import abspath, dirname, join
from socketserver import TCPServer
from http.server import SimpleHTTPRequestHandler


ROOT = join(dirname(abspath(__file__)), "html")
PORT = 7272


class DemoServer(TCPServer):
    allow_reuse_address = True

    def __init__(self, port=PORT):
        TCPServer.__init__(self, ("localhost", int(port)), SimpleHTTPRequestHandler)

    def serve(self, directory=ROOT):
        chdir(directory)
        print("Demo server starting on port %d." % self.server_address[1])
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            server.server_close()
        print("Demo server stopped.")


if __name__ == "__main__":
    import sys

    try:
        server = DemoServer(*sys.argv[1:])
    except (TypeError, ValueError):
        print(__doc__)
    else:
        server.serve()



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    import pandas as pd  # type: ignore
    from math import nan  # type: ignore
    import openpyxl  # type: ignore
except ImportError:
    raise ImportError(
        """Requirements (pandas, openpyxl) for XLSX support are not installed.
    Use 'pip install -U robotframework-datadriver[XLS]' to install XLSX support."""
    )
from robot.utils import is_truthy  # type: ignore

from .AbstractReaderClass import AbstractReaderClass


class xlsx_reader(AbstractReaderClass):
    def get_data_from_source(self):
        dtype = object if is_truthy(getattr(self, "preserve_xls_types", False)) else str
        data_frame = self.read_data_frame_from_file(dtype)
        self._analyse_header(list(data_frame))
        for row_index, row in enumerate(data_frame.values.tolist()):
            try:
                self._read_data_from_table(row)
            except Exception as e:
                e.row = row_index + 1
                raise e
        return self.data_table

    def read_data_frame_from_file(self, dtype):
        return pd.read_excel(
            self.file, sheet_name=self.sheet_name, dtype=dtype, engine="openpyxl"
        ).replace(nan, "", regex=True)



from DataDriver.AbstractReaderClass import AbstractReaderClass
from DataDriver.ReaderConfig import TestCaseData


class foo_reader(AbstractReaderClass):
    def get_data_from_source(self):
        return self._read_file_to_data_table()

    def _read_file_to_data_table(self):
        test_data = []
        flipflop = True
        for i in range(6):
            args = {'${fooarg}': i}
            tags = ["included"]
            if flipflop:
                tags = ["filtered"]
                flipflop = False
            else:
                flipflop = True
            if i == 3:
                tags = []
            test_data.append(TestCaseData(f"Test {i}", args, tags))
        return test_data



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from DataDriver.AbstractReaderClass import (
    AbstractReaderClass,
)  # inherit class from AbstractReaderClass
from DataDriver.ReaderConfig import TestCaseData  # return list of TestCaseData to DataDriver


class custom_reader(AbstractReaderClass):

    # This method will be called from DataDriver to get the TestCaseData list.
    def get_data_from_source(self):
        test_data = []
        for i in range(int(self.min), int(self.max)):  # Dummy code to just generate some data
            args = {
                "${var_1}": str(i),
                "${var_2}": str(i),
            }  # args is a dictionary. Variable name is the key, value is value.
            test_data.append(
                TestCaseData(f"test {i}", args, ["tag"])
            )  # add a TestCaseData object to the list of tests.
        return test_data  # return the list of TestCaseData to DataDriver



# Copyright 2018-  René Rohner
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
import os
import os.path
import re
import traceback
from glob import glob

from robot.api.logger import console  # type: ignore
from robot.libraries.BuiltIn import BuiltIn  # type: ignore
from robot.model.testsuite import TestSuite  # type: ignore
from robot.running.model import TestCase  # type: ignore
from robot.model.tags import Tags  # type: ignore
from robot.utils.dotdict import DotDict  # type: ignore
from robot.utils.importer import Importer  # type: ignore
from typing import Optional, Union, Any  # type: ignore

from .AbstractReaderClass import AbstractReaderClass  # type: ignore
from .ReaderConfig import ReaderConfig  # type: ignore
from .ReaderConfig import TestCaseData  # type: ignore
from .argument_utils import robot_options  # type: ignore
from .utils import (  # type: ignore
    PabotOpt,
    is_same_keyword,
    get_filter_dynamic_test_names,
    get_variable_value,
    is_pabot_dry_run,
    debug,
    warn,
    error,
    equally_partition_test_list,
    binary_partition_test_list,
    Encodings,
    TagHandling,
)

__version__ = "1.5.0"


class DataDriver:
    # region: docstring
    """

    ===================================================
    DataDriver for Robot Framework®
    ===================================================

    DataDriver is a Data-Driven extension for Robot Framework®.
    This document explains how to use the DataDriver library listener. For
    information about installation, support, and more, please visit the
    `project page <https://github.com/Snooz82/robotframework-datadriver>`_

    For more information about Robot Framework®, see http://robotframework.org.

    DataDriver is used/imported as Library but does not provide keywords
    which can be used in a test. DataDriver uses the Listener Interface
    Version 3 to manipulate the test cases and creates new test cases based
    on a Data-File that contains the data for Data-Driven Testing. These
    data file may be .csv , .xls or .xlsx files.

    Data Driver is also able to cooperate with Microsoft PICT. An Open
    Source Windows tool for data combination testing. Pict is able to
    generate data combinations based on textual model definitions.
    https://github.com/Microsoft/pict

    It is also possible to implement own DataReaders in Python to read
    your test data from some other sources, like databases or json files.


    Installation
    ------------

    If you already have Python >= 3.6 with pip installed, you can simply
    run:

    ``pip install --upgrade robotframework-datadriver``


    Excel Support
    ~~~~~~~~~~~~~

    For file support of ``xls`` or ``xlsx`` file you need to install the extra XLS or the dependencies.
    It contains the dependencies of pandas, numpy and xlrd. Just add [XLS] to your installation.
    New since version 3.6.

    ``pip install --upgrade robotframework-datadriver[XLS]``


    Python 2
    ~~~~~~~~

    or if you have Python 2 and 3 installed in parallel you may use

    ``pip3 install --upgrade robotframework-datadriver``

    DataDriver is compatible with Python 2.7 only in Version 0.2.7.

    ``pip install --upgrade robotframework-datadriver==0.2.7``

    Because Python 2.7 is deprecated, there are no new feature to python 2.7 compatible version.


    Table of contents
    -----------------

    -  `What DataDriver Does`_
    -  `How DataDriver Works`_
    -  `Usage`_
    -  `Structure of Test Suite`_
    -  `Structure of data file`_
    -  `Accessing Test Data From Robot Variables`_
    -  `Data Sources`_
    -  `File Encoding and CSV Dialect`_
    -  `Custom DataReader Classes`_
    -  `Selection of Test Cases to Execute`_
    -  `Configure DataDriver by Pre-Run Keyword`_
    -  `Pabot and DataDriver`_


    What DataDriver Does
    --------------------

    DataDriver is an alternative approach to create Data-Driven Tests with
    Robot Framework®. DataDriver creates multiple test cases based on a test
    template and data content of a csv or Excel file. All created tests
    share the same test sequence (keywords) and differ in the test data.
    Because these tests are created on runtime only the template has to be
    specified within the robot test specification and the used data are
    specified in an external data file.


    RoboCon 2020 Talk
    ~~~~~~~~~~~~~~~~~

    .. image:: https://img.youtube.com/vi/RtEUr1i4x3s/0.jpg
       :target: https://www.youtube.com/watch?v=RtEUr1i4x3s

    Brief overview what DataDriver is and how it works at the RoboCon 2020 in Helsiki.


    Alternative approach
    ~~~~~~~~~~~~~~~~~~~~

    DataDriver gives an alternative to the build in data driven approach
    like:

    .. code :: robotframework

        *** Settings ***
        Resource    login_resources.robot

        Suite Setup    Open my Browser
        Suite Teardown    Close Browsers
        Test Setup      Open Login Page
        Test Template    Invalid login


        *** Test Cases ***       User        Passwort
        Right user empty pass    demo        ${EMPTY}
        Right user wrong pass    demo        FooBar

        Empty user right pass    ${EMPTY}    mode
        Empty user empty pass    ${EMPTY}    ${EMPTY}
        Empty user wrong pass    ${EMPTY}    FooBar

        Wrong user right pass    FooBar      mode
        Wrong user empty pass    FooBar      ${EMPTY}
        Wrong user wrong pass    FooBar      FooBar

        *** Keywords ***
        Invalid login
            [Arguments]    ${username}    ${password}
            Input username    ${username}
            Input pwd    ${password}
            click login button
            Error page should be visible

    This inbuilt approach is fine for a hand full of data and a hand full of
    test cases. If you have generated or calculated data and specially if
    you have a variable amount of test case / combinations these robot files
    become quite a pain. With DataDriver you may write the same test case
    syntax but only once and deliver the data from en external data file.

    One of the rare reasons when Microsoft® Excel or LibreOffice Calc may be
    used in testing… ;-)

    `See example test suite <#example-suite>`__

    `See example csv table <#example-csv>`__


    How DataDriver Works
    --------------------

    When the DataDriver is used in a test suite it will be activated before
    the test suite starts. It uses the Listener Interface Version 3 of Robot
    Framework® to read and modify the test specification objects. After
    activation it searches for the ``Test Template`` -Keyword to analyze the
    ``[Arguments]`` it has. As a second step, it loads the data from the
    specified data source. Based on the ``Test Template`` -Keyword, DataDriver
    creates as much test cases as data sets are in the data source.

    In the case that data source is csv (Default)
    As values for the arguments of the ``Test Template`` -Keyword, DataDriver
    reads values from the column of the CSV file with the matching name of the
    ``[Arguments]``.
    For each line of the CSV data table, one test case will be created. It
    is also possible to specify test case names, tags and documentation for
    each test case in the specific test suite related CSV file.


    Usage
    -----

    Data Driver is a "Library Listener" but does not provide keywords.
    Because Data Driver is a listener and a library at the same time it
    sets itself as a listener when this library is imported into a test suite.

    To use it, just use it as Library in your suite. You may use the first
    argument (option) which may set the file name or path to the data file.

    Without any options set, it loads a .csv file which has the same name
    and path like the test suite .robot .



    **Example:**

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver
        Test Template    Invalid Logins

        *** Keywords ***
        Invalid Logins
            ...


    Structure of Test Suite
    -----------------------


    Requirements
    ~~~~~~~~~~~~

    In the Moment there are some requirements how a test
    suite must be structured so that the DataDriver can get all the
    information it needs.

     - only the first test case will be used as a template. All other test
       cases will be deleted.
     - Test cases have to be defined with a
       ``Test Template`` in Settings secion. Reason for this is,
       that the DataDriver needs to know the names of the test case arguments.
       Test cases do not have named arguments. Keywords do.
     - The keyword which is used as
       ``Test Template`` must be defined within the test suite (in the same
       \*.robot file). If the keyword which is used as ``Test Template`` is
       defined in a ``Resource`` the DataDriver has no access to its
       arguments names.


    Example Test Suite
    ~~~~~~~~~~~~~~~~~~

    .. code :: robotframework

        ***Settings***
        Library           DataDriver
        Resource          login_resources.robot
        Suite Setup       Open my Browser
        Suite Teardown    Close Browsers
        Test Setup        Open Login Page
        Test Template     Invalid Login

        *** Test Case ***
        Login with user ${username} and password ${password}    Default    UserData

        ***** *Keywords* *****
        Invalid login
            [Arguments]    ${username}    ${password}
            Input username    ${username}
            Input pwd    ${password}
            click login button
            Error page should be visible

    In this example, the DataDriver is activated by using it as a Library.
    It is used with default settings.
    As ``Test Template`` the keyword ``Invalid Login`` is used. This
    keyword has two arguments. Argument names are ``${username}`` and
    ``${password}``. These names have to be in the CSV file as column
    header. The test case has two variable names included in its name,
    which does not have any functionality in Robot Framework®. However, the
    Data Driver will use the test case name as a template name and
    replaces the variables with the specific value of the single generated
    test case.
    This template test will only be used as a template. The specified data
    ``Default`` and ``UserData`` would only be used if no CSV file has
    been found.


    Structure of data file
    ----------------------


    min. required columns
    ~~~~~~~~~~~~~~~~~~~~~

    -  ``*** Test Cases ***`` column has to be the first one.
    -  *Argument columns:* For each argument of the ``Test Template``
       keyword one column must be existing in the data file as data source.
       The name of this column must match the variable name and syntax.


    optional columns
    ~~~~~~~~~~~~~~~~

    -  *[Tags]* column may be used to add specific tags to a test case. Tags
       may be comma separated.
    -  *[Documentation]* column may be used to add specific test case
       documentation.


    Example Data file
    ~~~~~~~~~~~~~~~~~

    +-------------+-------------+-------------+-------------+------------------+
    | \**\* Test  | ${username} | ${password} | [Tags]      | [Documentation]  |
    | Cases \**\* |             |             |             |                  |
    |             |             |             |             |                  |
    +=============+=============+=============+=============+==================+
    | Right user  | demo        | ${EMPTY}    | 1           | This is a test   |
    | empty pass  |             |             |             | case             |
    |             |             |             |             | documentation of |
    |             |             |             |             | the first one.   |
    +-------------+-------------+-------------+-------------+------------------+
    | Right user  | demo        | FooBar      | 2,3,foo     | This test        |
    | wrong pass  |             |             |             | case has         |
    |             |             |             |             | the Tags         |
    |             |             |             |             | 2,3 and foo      |
    |             |             |             |             | assigned.        |
    +-------------+-------------+-------------+-------------+------------------+
    |             | ${EMPTY}    | mode        | 1,2,3,4     | This test        |
    |             |             |             |             | case has a       |
    |             |             |             |             | generated        |
    |             |             |             |             | name based       |
    |             |             |             |             | on template      |
    |             |             |             |             | name.            |
    +-------------+-------------+-------------+-------------+------------------+
    |             | ${EMPTY}    | ${EMPTY}    |             |                  |
    +-------------+-------------+-------------+-------------+------------------+
    |             | ${EMPTY}    | FooBar      |             |                  |
    +-------------+-------------+-------------+-------------+------------------+
    |             | FooBar      | mode        |             |                  |
    +-------------+-------------+-------------+-------------+------------------+
    |             | FooBar      | ${EMPTY}    |             |                  |
    +-------------+-------------+-------------+-------------+------------------+
    |             | FooBar      | FooBar      |             |                  |
    +-------------+-------------+-------------+-------------+------------------+

    In this data file, eight test cases are defined. Each line specifies one
    test case. The first two test cases have specific names. The other six
    test cases will generate names based on template test cases name with
    the replacement of variables in this name. The order of columns is
    irrelevant except the first column, ``*** Test Cases ***``

    Supported Data Types
    ~~~~~~~~~~~~~~~~~~~~

    In general DataDriver supports any Object that is handed over from the DataReader.
    However the text based readers for csv, excel and so do support different types as well.
    DataDriver supports Robot Framework® Scalar variables as well as Dictionaries and Lists.
    It also support python literal evaluations.

    Scalar Variables
    ^^^^^^^^^^^^^^^^

    The Prefix ``$`` defines that the value in the cell is taken as in Robot Framework® Syntax.
    ``String`` is ``str``, ``${1}`` is ``int`` and ``${None}`` is NoneType.
    The Prefix only defines the value typ. It can also be used to assign a scalar to a dictionary key.
    See example table: ``${user}[id]``


    Dictionary Variables
    ^^^^^^^^^^^^^^^^^^^^

    Dictionaries can be created in different ways.

    One option is, to use the prefix ``&``.
    If a variable is defined that was (i.e. ``&{dict}``) the cell value is interpreted the same way,
    the BuiltIn keyword `Create Dictionary <https://robotframework.org/robotframework/latest/libraries/BuiltIn.html#Create%20Dictionary>`_ would do.
    The arguments here are comma (``,``) separated.
    See example table: ``&{dict}``

    The other option is to define scalar variables in dictionary syntax like ``${user}[name]`` or ``${user.name}``
    That can be also nested dictionaries. DataDriver will create Robot Framework® (DotDict) Dictionaries, that can be accessed with ``${user.name.first}``
    See example table: ``${user}[name][first]``


    List Variables
    ^^^^^^^^^^^^^^

    Lists can be created with the prefix ``@`` as comma (``,``) separated list.
    See example table: ``@{list}``

    Be aware that a list with an empty string has to be the cell content `${Empty}`.

    Python Literals
    ^^^^^^^^^^^^^^^

    DataDriver can evaluate Literals.
    It uses the prefix ``e`` for that. (i.e. ``e{list_eval}``)
    For that it uses `BuiltIn Evaluate <https://robotframework.org/robotframework/latest/libraries/BuiltIn.html#Evaluate>`_

    See example table: ``e{user.chk}``


    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+
    |  ``*** Test Cases ***``  |  ``${scalar}``        |  ``@{list}``  |  ``e{list_eval}``       |  ``&{dict}``                |  ``e{dict_eval}``                        |  ``e{eval}``             |  ``${exp_eval}``  |  ``${user}[id]``  |  ``${user}[name][first]``  |  ``${user.name.last}``  |  ``e{user.chk}``                                                 |
    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+
    |  ``One``                 |  ``Sum List``         |  ``1,2,3,4``  |  ``["1","2","3","4"]``  |  ``key=value``              |  ``{'key': 'value'}``                    |  ``[1,2,3,4]``           |  ``10``           |  ``1``            |  ``Pekka``                 |  ``Klärck``             |  ``{'id': '1', 'name': {'first': 'Pekka', 'last': 'Klärck'}}``   |
    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+
    |  ``Two``                 |  ``Should be Equal``  |  ``a,b,c,d``  |  ``["a","b","c","d"]``  |  ``key,value``              |  ``{'key': 'value'}``                    |  ``True``                |  ``${true}``      |  ``2``            |  ``Ed``                    |  ``Manlove``            |  ``{'id': '2', 'name': {'first': 'Ed', 'last': 'Manlove'}}``     |
    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+
    |  ``Three``               |  ``Whos your Daddy``  |  ``!,",',$``  |  ``["!",'"',"'","$"]``  |  ``z,value,a,value2``       |  ``{'a': 'value2', 'z': 'value'}``       |  ``{'Daddy' : 'René'}``  |  ``René``         |  ``3``            |  ``Tatu``                  |  ``Aalto``              |  ``{'id': '3', 'name': {'first': 'Tatu', 'last': 'Aalto'}}``     |
    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+
    |  ``4``                   |  ``Should be Equal``  |  ``1``        |  ``["1"]``              |  ``key=value``              |  ``{'key': 'value'}``                    |  ``1``                   |  ``${1}``         |  ``4``            |  ``Jani``                  |  ``Mikkonen``           |  ``{'id': '4', 'name': {'first': 'Jani', 'last': 'Mikkonen'}}``  |
    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+
    |  ``5``                   |  ``Should be Equal``  |               |  ``[]``                 |  ``a=${2}``                 |  ``{'a':2}``                             |  ``"string"``            |  ``string``       |  ``5``            |  ``Mikko``                 |  ``Korpela``            |  ``{'id': '5', 'name': {'first': 'Mikko', 'last': 'Korpela'}}``  |
    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+
    |  ``6``                   |  ``Should be Equal``  |  ``[1,2]``    |  ``["[1","2]"]``        |  ``key=value,key2=value2``  |  ``{'key': 'value', 'key2': 'value2'}``  |  ``None``                |  ``${none}``      |  ``6``            |  ``Ismo``                  |  ``Aro``                | ``{'id': '6', 'name': {'first': 'Ismo', 'last': 'Aro'}}``        |
    +--------------------------+-----------------------+---------------+-------------------------+-----------------------------+------------------------------------------+--------------------------+-------------------+-------------------+----------------------------+-------------------------+------------------------------------------------------------------+


    Accessing Test Data From Robot Variables
    ----------------------------------------

    If neccesary it is possible to access the fetched data tables directly from a Robot Framework® variable.
    This could be helpfull in Test Setup or in Suite Setup.

    There are three variables available within the Data-Driven Suite:

    @{DataDriver_DATA_LIST}
    ~~~~~~~~~~~~~~~~~~~~~~~

    A list as suite variable containing a robot dictionary for each test case that is selected for execution.

    .. code :: json

        [
          {
            "test_case_name": "Right user empty pass",
            "arguments": {
              "${username}": "demo",
              "${password}": "${EMPTY}"
            },
            "tags": [
              "1"
            ],
            "documentation": "This is a test case documentation of the first one."
          },
          {
            "test_case_name": "Right user wrong pass",
            "arguments": {
              "${username}": "demo",
              "${password}": "FooBar"
            },
            "tags": [
              "2",
              "3",
              "foo"
            ],
            "documentation": "This test case has the Tags 2,3 and foo"
          },
          {
            "test_case_name": "Login with user '${EMPTY}' and password 'mode'",
            "arguments": {
              "${username}": "${EMPTY}",
              "${password}": "mode"
            },
            "tags": [
              "1",
              "2",
              "3",
              "4"
            ],
            "documentation": "This test case has a generated name based on template name."
          },
          {
            "test_case_name": "Login with user '${EMPTY}' and password '${EMPTY}'",
            "arguments": {
              "${username}": "${EMPTY}",
              "${password}": "${EMPTY}"
            },
            "tags": [
              ""
            ],
            "documentation": ""
          },
          {
            "test_case_name": "Login with user '${EMPTY}' and password 'FooBar'",
            "arguments": {
              "${username}": "${EMPTY}",
              "${password}": "FooBar"
            },
            "tags": [
              ""
            ],
            "documentation": ""
          },
          {
            "test_case_name": "Login with user 'FooBar' and password 'mode'",
            "arguments": {
              "${username}": "FooBar",
              "${password}": "mode"
            },
            "tags": [
              "foo",
              "1"
            ],
            "documentation": ""
          },
          {
            "test_case_name": "Login with user 'FooBar' and password '${EMPTY}'",
            "arguments": {
              "${username}": "FooBar",
              "${password}": "${EMPTY}"
            },
            "tags": [
              "foo"
            ],
            "documentation": ""
          },
          {
            "test_case_name": "Login with user 'FooBar' and password 'FooBar'",
            "arguments": {
              "${username}": "FooBar",
              "${password}": "FooBar"
            },
            "tags": [
              "foo",
              "2"
            ],
            "documentation": ""
          }
        ]

    This can be accessed as usual in Robot Framework®.

    ``${DataDriver_DATA_LIST}[2][arguments][\${password}]`` would result in ``mode`` .



    &{DataDriver_DATA_DICT}
    ~~~~~~~~~~~~~~~~~~~~~~~

    A dictionary as suite variable that contains the same data as the list, with the test names as keys.

    .. code :: json

        {
          "Right user empty pass": {
            "test_case_name": "Right user empty pass",
            "arguments": {
              "${username}": "demo",
              "${password}": "${EMPTY}"
            },
            "tags": [
              "1"
            ],
            "documentation": "This is a test case documentation of the first one."
          },
          "Right user wrong pass": {
            "test_case_name": "Right user wrong pass",
            "arguments": {
              "${username}": "demo",
              "${password}": "FooBar"
            },
            "tags": [
              "2",
              "3",
              "foo"
            ],
            "documentation": "This test case has the Tags 2,3 and foo"
          },
          "Login with user '${EMPTY}' and password 'mode'": {
            "test_case_name": "Login with user '${EMPTY}' and password 'mode'",
            "arguments": {
              "${username}": "${EMPTY}",
              "${password}": "mode"
            },
            "tags": [
              "1",
              "2",
              "3",
              "4"
            ],
            "documentation": "This test case has a generated name based on template name."
          },
          "Login with user '${EMPTY}' and password '${EMPTY}'": {
            "test_case_name": "Login with user '${EMPTY}' and password '${EMPTY}'",
            "arguments": {
              "${username}": "${EMPTY}",
              "${password}": "${EMPTY}"
            },
            "tags": [
              ""
            ],
            "documentation": ""
          },
          "Login with user '${EMPTY}' and password 'FooBar'": {
            "test_case_name": "Login with user '${EMPTY}' and password 'FooBar'",
            "arguments": {
              "${username}": "${EMPTY}",
              "${password}": "FooBar"
            },
            "tags": [
              ""
            ],
            "documentation": ""
          },
          "Login with user 'FooBar' and password 'mode'": {
            "test_case_name": "Login with user 'FooBar' and password 'mode'",
            "arguments": {
              "${username}": "FooBar",
              "${password}": "mode"
            },
            "tags": [
              "foo",
              "1"
            ],
            "documentation": ""
          },
          "Login with user 'FooBar' and password '${EMPTY}'": {
            "test_case_name": "Login with user 'FooBar' and password '${EMPTY}'",
            "arguments": {
              "${username}": "FooBar",
              "${password}": "${EMPTY}"
            },
            "tags": [
              "foo"
            ],
            "documentation": ""
          },
          "Login with user 'FooBar' and password 'FooBar'": {
            "test_case_name": "Login with user 'FooBar' and password 'FooBar'",
            "arguments": {
              "${username}": "FooBar",
              "${password}": "FooBar"
            },
            "tags": [
              "foo",
              "2"
            ],
            "documentation": ""
          }
        }

    &{DataDriver_TEST_DATA}
    ~~~~~~~~~~~~~~~~~~~~~~~

    A dictionary as test variable that contains the test data of the current test case.
    This dictionary does also contain arguments that are not used in the ``Test Template`` keyword.
    This can be used in Test Setup and within a test case.

    .. code :: json

        {
          "test_case_name": "Right user wrong pass",
          "arguments": {
            "${username}": "demo",
            "${password}": "FooBar"
          },
          "tags": [
            "2",
            "3",
            "foo"
          ],
          "documentation": "This test case has the Tags 2,3 and foo"
        }


    Data Sources
    ------------


    CSV / TSV (Character-separated values)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    By default DataDriver reads csv files. With the `Encoding and CSV
    Dialect <#file-encoding-and-csv-dialect>`__ settings you may configure which
    structure your data source has.


    XLS / XLSX Files
    ~~~~~~~~~~~~~~~~

    To use Excel file types, you have to install DataDriver with the Extra XLS.

    If you want to use Excel based data sources, you may just set the file
    to the extention or you may point to the correct file. If the extention
    is ".xls" or ".xlsx" DataDriver will interpret it as Excel file.
    You may select the sheet which will be read by the option ``sheet_name``.
    By default it is set to 0 which will be the first table sheet.
    You may use sheet index (0 is first sheet) or sheet name(case sensitive).
    XLS interpreter will ignore all other options like encoding, delimiters etc.

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    .xlsx

    or:

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    file=my_data_source.xlsx    sheet_name=2nd Sheet


    MS Excel and typed cells
    ^^^^^^^^^^^^^^^^^^^^^^^^

    Microsoft Excel xls or xlsx file have the possibility to type thair data
    cells. Numbers are typically of the type float. If these data are not
    explicitly defined as text in Excel, pandas will read it as the type
    that is has in excel. Because we have to work with strings in Robot
    Framework® these data are converted to string. This leads to the
    situation that a European time value like "04.02.2019" (4th January
    2019) is handed over to Robot Framework® in Iso time "2019-01-04
    00:00:00". This may cause unwanted behavior. To mitigate this risk you
    should define Excel based files explicitly as text within Excel.

    Alternatively you may deactivate that string conversion.
    To do so, you have to add the option ``preserve_xls_types`` to ``True``.
    In that case, you will get str, float, boolean, int, datetime.time,
    datetime.datetime and some others.

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    file=my_data_source.xlsx    preserve_xls_types=True

    PICT (Pairwise Independent Combinatorial Testing)
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Pict is able to generate data files based on a model file.
    https://github.com/Microsoft/pict

    Documentation: https://github.com/Microsoft/pict/blob/master/doc/pict.md


    Requirements
    ^^^^^^^^^^^^

    -  Path to pict.exe must be set in the %PATH% environment variable.
    -  Data model file has the file extention ".pict"
    -  Pict model file must be encoded in UTF-8


    How it works
    ^^^^^^^^^^^^

    If the file option is set to a file with the extention pict, DataDriver
    will hand over this file to pict.exe and let it automatically generates
    a file with the extention ".pictout". This file will the be used as data
    source for the test generation. (It is tab seperated and UTF-8 encoded)
    Except the file option all other options of the library will be ignored.

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    my_model_file.pict


    Glob File Pattern
    ~~~~~~~~~~~~~~~~~

    This module implements a reader class that creates a test case for each file or folder that matches the given glob pattern.

    With an optional argument "arg_name" you can modify the argument that will be set. See folder example.

    Example with json files:

    .. code :: robotframework

        *** Settings ***
        Library           DataDriver    file=${CURDIR}/DataFiles/*_File.json    reader_class=glob_reader
        Library           OperatingSystem
        Test Template     Test all Files


        *** Test Cases ***
        Glob_Reader_Test    Wrong_File.NoJson


        *** Keywords ***
        Test all Files
            [Arguments]    ${file_name}
            ${file_content}=    Get File    ${file_name}
            ${content}=    Evaluate    json.loads($file_content)["test_case"]
            Should Be Equal    ${TEST_NAME}    ${content}


    Example with folders:

    .. code :: robotframework

        *** Settings ***
        Library           DataDriver    file=${CURDIR}/FoldersToFind/*/    reader_class=glob_reader    arg_name=\${folder_name}
        Library           OperatingSystem
        Test Template     Test all Files


        *** Test Cases ***
        Glob_Reader_Test    Wrong_File.NoJson


        *** Keywords ***
        Test all Files
            [Arguments]    ${folder_name}
            ${content}=    Get File    ${folder_name}/verify.txt
            Should Be Equal    ${TEST_NAME}    ${content}


    File Encoding and CSV Dialect
    -----------------------------

    CSV is far away from well designed and has absolutely no "common"
    format. Therefore it is possible to define your own dialect or use
    predefined. The default is Excel-EU which is a semicolon separated
    file.
    These Settings are changeable as options of the Data Driver Library.


    file=
    ~~~~~

    .. code :: robotframework

        *** Settings ***
        Library         DataDriver    file=../data/my_data_source.csv


    -  None(default): Data Driver will search in the test suites folder if a
       \*.csv file with the same name than the test suite \*.robot file exists
    -  only file extention: if you just set a file extentions like ".xls" or
       ".xlsx" DataDriver will search
    -  absolute path: If an absolute path to a file is set, DataDriver tries
       to find and open the given data file.
    -  relative path: If the option does not point to a data file as an
       absolute path, Data Driver tries to find a data file relative to the
       folder where the test suite is located.


    encoding=
    ~~~~~~~~~

    ``encoding=`` must be set if it shall not be cp1252.

    **Examples**:

    ``cp1252, ascii, iso-8859-1, latin-1, utf_8, utf_16, utf_16_be, utf_16_le``

    **cp1252** is:

    - Code Page 1252
    - Windows-1252
    - Windows Western European

    Most characters are same between ISO-8859-1 (Latin-1) except for the code points 128-159 (0x80-0x9F).
    These Characters are available in cp1252 which are not present in Latin-1.

    ``€ ‚ ƒ „ … † ‡ ˆ ‰ Š ‹ Œ Ž ‘ ’ “ ” • – — ˜ ™ š › œ ž Ÿ``

    See `Python Standard Encoding <https://docs.python.org/3/library/codecs.html#standard-encodings>`_ for more encodings


    dialect=
    ~~~~~~~~

    You may change the CSV Dialect here.
    The dialect option can be one of the following:
    - Excel-EU
    - excel
    - excel-tab
    - unix
    - UserDefined

    supported Dialects are:

    .. code:: python

        "Excel-EU"
            delimiter=';',
            quotechar='"',
            escapechar='\\',
            doublequote=True,
            skipinitialspace=False,
            lineterminator="\\r\\n",
            quoting=csv.QUOTE_ALL

        "excel"
            delimiter = ','
            quotechar = '"'
            doublequote = True
            skipinitialspace = False
            lineterminator = '\\r\\n'
            quoting = QUOTE_MINIMAL

        "excel-tab"
            delimiter = '\\t'
            quotechar = '"'
            doublequote = True
            skipinitialspace = False
            lineterminator = '\\r\\n'
            quoting = QUOTE_MINIMAL

        "unix"
            delimiter = ','
            quotechar = '"'
            doublequote = True
            skipinitialspace = False
            lineterminator = '\\n'
            quoting = QUOTE_ALL




    Usage in Robot Framework®

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    my_data_file.csv    dialect=excel



    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    my_data_file.csv    dialect=excel_tab



    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    my_data_file.csv    dialect=unix_dialect



    Example User Defined
    ^^^^^^^^^^^^^^^^^^^^

    User may define the format completely free.
    If an option is not set, the default values are used.
    To register a userdefined format user have to set the
    option ``dialect`` to ``UserDefined``


    Usage in Robot Framework®

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    my_data_file.csv
        ...    dialect=UserDefined
        ...    delimiter=.
        ...    lineterminator=\\n




    Defaults:
    ~~~~~~~~~

    .. code:: python

        file=None,
        encoding='cp1252',
        dialect='Excel-EU',
        delimiter=';',
        quotechar='"',
        escapechar='\\\\',
        doublequote=True,
        skipinitialspace=False,
        lineterminator='\\r\\n',
        sheet_name=0


    Custom DataReader Classes
    -------------------------

    It is possible to write your own DataReader Class as a plugin for DataDriver.
    DataReader Classes are called from DataDriver to return a list of TestCaseData.


    Using Custom DataReader
    ~~~~~~~~~~~~~~~~~~~~~~~

    DataReader classes are loaded dynamically into DataDriver while runtime.
    DataDriver identifies the DataReader to load by the file extantion of the data file or by the option ``reader_class``.


    Select Reader by File Extension:
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    file=mydata.csv

    This will load the class ``csv_reader`` from ``csv_reader.py`` from the same folder.


    Select Reader by Option:
    ^^^^^^^^^^^^^^^^^^^^^^^^

    .. code :: robotframework

        *** Settings ***
            Library    DataDriver   file=mydata.csv    reader_class=generic_csv_reader    dialect=userdefined   delimiter=\\t    encoding=UTF-8

    This will load the class ``generic_csv_reader`` from ``generic_csv_reader.py`` from same folder.


    Create Custom Reader
    ~~~~~~~~~~~~~~~~~~~~

    Recommendation:

    Have a look to the Source Code of existing DataReader like ``csv_reader.py`` or ``generic_csv_reader.py`` .

    To write your own reader, create a class inherited from ``AbstractReaderClass``.

    Your class will get all available configs from DataDriver as an object of ``ReaderConfig`` on ``__init__``.

    DataDriver will call the method ``get_data_from_source``
    This method should then load your data from your custom source and stores them into list of object of ``TestCaseData``.
    This List of ``TestCaseData`` will be returned to DataDriver.

    ``AbstractReaderClass`` has also some optional helper methods that may be useful.

    You can either place the custom reader with the others in DataDriver folder or anywhere on the disk.
    In the first case or if your custom reader is in python path just use it like the others by name:

    .. code :: robotframework

        *** Settings ***
        Library          DataDriver    reader_class=my_reader

    In case it is somewhere on the disk, it is possible to use an absolute or relative path to a custom Reader.
    Imports of custom readers follow the same rules like importing Robot Framework® libraries.
    Path can be relative to ${EXECDIR} or to DataDriver/__init__.py:


    .. code :: robotframework

        *** Settings ***
        Library          DataDriver    reader_class=C:/data/my_reader.py    # set custom reader
        ...                            file_search_strategy=None            # set DataDriver to not check file
        ...                            min=0                                # kwargs arguments for custom reader
        ...                            max=62

    This `my_reader.py` should implement a class inherited from AbstractReaderClass that is named `my_reader`.

    .. code :: python

        from DataDriver.AbstractReaderClass import AbstractReaderClass  # inherit class from AbstractReaderClass
        from DataDriver.ReaderConfig import TestCaseData  # return list of TestCaseData to DataDriver


        class my_reader(AbstractReaderClass):

            def get_data_from_source(self):  # This method will be called from DataDriver to get the TestCaseData list.
                test_data = []
                for i in range(int(self.kwargs['min']), int(self.kwargs['max'])):  # Dummy code to just generate some data
                    args = {'${var_1}': str(i), '${var_2}': str(i)}  # args is a dictionary. Variable name is the key, value is value.
                    test_data.append(TestCaseData(f'test {i}', args, ['tag']))  # add a TestCaseData object to the list of tests.
                return test_data  # return the list of TestCaseData to DataDriver


    See other readers as example.


    Selection of Test Cases to Execute
    ----------------------------------

    Because test cases that are created by DataDriver after parsing while execution,
    it is not possible to use some Robot Framework® methods to select test cases.


    Examples for options that have to be used differently:

    +-------------------+-----------------------------------------------------------------------+
    | robot option      | Description                                                           |
    +===================+=======================================================================+
    | ``--test``        | Selects the test cases by name.                                       |
    +-------------------+-----------------------------------------------------------------------+
    | ``--task``        | Alias for --test that can be used when executing tasks.               |
    +-------------------+-----------------------------------------------------------------------+
    | ``--rerunfailed`` | Selects failed tests from an earlier output file to be re-executed.   |
    +-------------------+-----------------------------------------------------------------------+
    | ``--include``     | Selects the test cases by tag.                                        |
    +-------------------+-----------------------------------------------------------------------+
    | ``--exclude``     | Selects the test cases by tag.                                        |
    +-------------------+-----------------------------------------------------------------------+


    Selection of test cases by name
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    Select a single test case:
    ^^^^^^^^^^^^^^^^^^^^^^^^^^

    To execute just a single test case by its exact name it is possible to execute the test suite
    and set the global variable ${DYNAMICTEST} with the name of the test case to execute as value.
    Pattern must be ``suitename.testcasename``.

    Example:

    .. code ::

        robot --variable "DYNAMICTEST:my suite name.test case to be executed" my_suite_name.robot

    Pabot uses this feature to execute a single test case when using ``--testlevelsplit``


    Select a list of test cases:
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    It is possible to set a list of test case names by using the variable ${DYNAMICTESTS} (plural).
    This variable must be a string and the list of names must be pipe-seperated (``|``).

    Example:

    .. code::

        robot --variable DYNAMICTESTS:firstsuitename.testcase1|firstsuitename.testcase3|anothersuitename.othertestcase foldername

    It is also possible to set the variable @{DYNAMICTESTS} as a list variable from i.e. python code.


    Re-run failed test cases:
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Because it is not possible to use the command line argument ``--rerunfailed`` from robot directly,
    DataDriver brings a Pre-Run-Modifier that handles this issue.

    Normally reexecution of failed testcases has three steps.

    - original execution
    - re-execution the failed ones based on original execution output
    - merging original execution output with re-execution output

    The DataDriver.rerunfailed Pre-Run-Modifier removes all passed test cases based on a former output.xml.

    Example:

    .. code ::

        robot --output original.xml tests                                                    # first execute all tests
        robot --prerunmodifier DataDriver.rerunfailed:original.xml --output rerun.xml tests  # then re-execute failing
        rebot --merge original.xml rerun.xml                                                 # finally merge results


    Be aware, that in this case it is not allowed to use "``:``" as character in the original output file path.
    If you want to set a full path on windows like ``e:\\myrobottest\\output.xml`` you have to use "``;``"
    as argument seperator.

    Example:

    .. code ::

        robot --prerunmodifier DataDriver.rerunfailed;e:\\myrobottest\\output.xml --output e:\\myrobottest\\rerun.xml tests



    Filtering with tags.
    ~~~~~~~~~~~~~~~~~~~~

    New in ``0.3.1``

    It is possible to use tags to filter the data source.
    To use this, tags must be assigned to the test cases in data source.


    Robot Framework® Command Line Arguments
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    To filter the source, the normal command line arguments of Robot Framework® can be used.
    See Robot Framework® Userguide_ for more information
    Be aware that the filtering of Robot Framework® itself is done before DataDriver is called.
    This means if the Template test is already filtered out by Robot Framework®, DataDriver can never be called.
    If you want to use ``--include`` the DataDriver TestSuite should have a ``DefaultTag`` or ``ForceTag`` that
    fulfills these requirements.

    .. _Userguide: http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#tag-patterns

    Example: ``robot --include 1OR2 --exclude foo DataDriven.robot``


    Filter based on Library Options
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    It is also possible to filter the data source by an init option of DataDriver.
    If these Options are set, Robot Framework® Filtering will be ignored.

    Example:

    .. code :: robotframework

        *** Settings ***
        Library    DataDriver    include=1OR2    exclude=foo




    Configure DataDriver by Pre-Run Keyword
    ---------------------------------------

    With ``config_keyword=`` it's possible to name a keyword that will be called from Data Driver before it starts the actual processing of the ``data file``.
    One possible usage is if the ``data file`` itself shall be created by another keyword dynamically during the execution of the Data Driver test suite.
    The ``config_keyword=`` can be used to call that keyword and return the updated arguments (e.g. ``file``) back to the Data Driver Library.

    The ``config keyword``

    - May be defined globally or inside each testsuite individually
    - Gets all the arguments, that Data Driver gets from Library import, as a Robot Dictionary
    - Shall return the (updated) Data Driver arguments as a Robot Dictionary

    Usage in Robot Framework®

    .. code :: robotframework

        *** Settings ***
        Library           OperatingSystem
        Library           DataDriver    dialect=excel    encoding=utf_8   config_keyword=Config
        Test Template     The Keyword

        *** Test Cases ***
        Test    aaa

        *** Keywords ***
        The Keyword
            [Arguments]    ${var}
            Log To Console    ${var}

        Config
            [Arguments]    ${original_config}
            Log To Console    ${original_config.dialect}                # just a log of the original
            Create File    ${CURDIR}/test321.csv
            ...    *** Test Cases ***,\${var},\\n123,111,\\n321,222,      # generating file
            ${new_config}=    Create Dictionary    file=test321.csv     # set file attribute in a dictionary
            [Return]    ${new_config}                                   # returns {'file': 'test321.csv'}



    Pabot and DataDriver
    --------------------

    You should use Pabot version 1.10.0 or newer.

    DataDriver supports ``--testlevelsplit`` from pabot only if the PabotLib is in use.
    Use ``--pabotlib`` to enable that.

    When using pabot like this, DataDriver automatically splits the amount of test cases into nearly same sized groups.
    Is uses the processes count from pabot to calculate the groups.
    When using 8 processes with 100 test cases you will get 8 groups of tests with the size of 12 to 13 tests.
    These 8 groups are then executed as one block with 8 processes.
    This reduces a lot of overhead with Suite Setup and Teardown.

    You can switch between three modes:

    - ``Equal``: means it creates equal sizes groups
    - ``Binary``: is more complex. it created a decreasing size of containers to support better balancing.
    - ``Atomic``: it does not group tests at all and runs really each test case in a separate thread.

    This can be set by ``optimize_pabot`` in Library import.


    **Example**:

    .. code :: robotframework

        *** Settings ***
        Library          DataDriver    optimize_pabot=Binary

    Binary creates with 40 test cases and 8 threads something like that:

    .. code ::

        P01: 01,02,03,04,05
        P02: 06,07,08,09,10
        P03: 11,12,13,14,15
        P04: 16,17,18,19,20
        P05: 21,22,23
        P06: 24,25,26
        P07: 27,28,29
        P08: 30,31,32
        P09: 33
        P10: 34
        P11: 35
        P12: 36
        P13: 37
        P14: 38
        P15: 39
        P16: 40

    """
    # endregion

    ROBOT_LIBRARY_DOC_FORMAT = "reST"
    ROBOT_LIBRARY_VERSION = __version__
    ROBOT_LISTENER_API_VERSION = 3
    ROBOT_LIBRARY_SCOPE = "TEST SUITE"

    def __init__(
        self,
        file: Optional[str] = None,
        encoding: Union[Encodings, Any] = Encodings.cp1252,
        dialect: str = "Excel-EU",
        delimiter: str = ";",
        quotechar: str = '"',
        escapechar: str = "\\",
        doublequote: bool = True,
        skipinitialspace: bool = False,
        lineterminator: str = "\r\n",
        *,
        sheet_name: Union[str, int] = 0,
        reader_class: Optional[str] = None,
        file_search_strategy: str = "PATH",
        file_regex: str = r"(?i)(.*?)(\.csv)",
        include: Optional[str] = None,
        exclude: Optional[str] = None,
        handle_template_tags: TagHandling = TagHandling.UnsetTags,
        listseperator: str = ",",
        config_keyword: Optional[str] = None,
        optimize_pabot: PabotOpt = PabotOpt.Equal,
        **kwargs: Any,
    ):
        # region: docstring
        """**Example:**

.. code :: robotframework

    *** Settings ***
    Library    DataDriver

Options
~~~~~~~

.. code :: robotframework

    *** Settings ***
    Library    DataDriver
    ...    file=None
    ...    encoding=cp1252
    ...    dialect=Excel-EU
    ...    delimiter=;
    ...    quotechar="
    ...    escapechar=\\\\
    ...    doublequote=True
    ...    skipinitialspace=False
    ...    lineterminator=\\r\\n
    ...    *  # the following are named only arguments
    ...    sheet_name=0
    ...    reader_class=None
    ...    file_search_strategy=PATH
    ...    file_regex=(?i)(.*?)(\\.csv)
    ...    include=None
    ...    exclude=None
    ...    handle_template_tags=UnsetTags
    ...    listseperator=,
    ...    config_keyword=None
    ...    optimize_pabot=Equal
    ...    &{kwargs}

File
^^^^

Defines which data file to be read. May be None or an extension like ``.txt`` or relative or absolute path.
``file_search_strategy=`` and ``file_regex`` may also have influence here.


Encoding
^^^^^^^^

``encoding`` must be set if it shall not be cp1252.

**Examples**:

``cp1252, ascii, iso-8859-1, latin-1, utf_8, utf_16, utf_16_be, utf_16_le``

See `Python Standard Encoding <https://docs.python.org/3/library/codecs.html#standard-encodings>`_ for more encodings


Dialect to LineTerminator
^^^^^^^^^^^^^^^^^^^^^^^^^

Defines how a csv file is interpreted.


Reader Class
^^^^^^^^^^^^

Defines which custom DataReader shall be loaded and get handed over all configs to deliver the TestCaseData.


Include & Exclude
^^^^^^^^^^^^^^^^^

Alternative way to CLI Options of Robot Framework® to filter the specific TestCaseData based on given Tags.


Handle Template Tags
^^^^^^^^^^^^^^^^^^^^

It can be configured how the tags of the template test will be added to the data-driven test cases.


List Separator
^^^^^^^^^^^^^^

Defines how list values or dictionary values are seperated.


Config Keyword
^^^^^^^^^^^^^^

Keyword that is executed before DataDriver starts working. Can manipulate all data.


Optimize Pabot
^^^^^^^^^^^^^^

When DataDriver is used together with Pabot, it optimizes the ``--testlevelsplit`` to be faster.

        """
        # endregion
        self.ROBOT_LIBRARY_LISTENER = self
        try:
            re.compile(file_regex)
        except re.error as e:
            file_regex = r"(?i)(.*?)(\.csv)"
            console(f"[ DataDriver ] invalid Regex! used {file_regex} instead.")
            console(e)

        self.robot_options = robot_options()
        self.include = self.robot_options["include"] if not include else include
        self.exclude = self.robot_options["exclude"] if not exclude else exclude
        self.handle_template_tags = handle_template_tags
        encoding_str = encoding.name if isinstance(encoding, Encodings) else str(encoding)

        self.config_dict = DotDict(
            file=file,
            encoding=encoding_str,
            dialect=dialect,
            delimiter=delimiter,
            quotechar=quotechar,
            escapechar=escapechar,
            doublequote=doublequote,
            skipinitialspace=skipinitialspace,
            lineterminator=lineterminator,
            sheet_name=sheet_name,
            reader_class=reader_class,
            file_search_strategy=file_search_strategy.upper(),
            file_regex=file_regex,
            include=self.include,
            exclude=self.exclude,
            handle_template_tags=self.handle_template_tags,
            list_separator=listseperator,
            config_keyword=config_keyword,
            optimize_pabot=optimize_pabot,
            **kwargs,
        )

        self.reader_config = ReaderConfig(**self.config_dict)
        self.suite_name = None
        self.suite_source = None
        self.template_test = None
        self.template_keyword = None
        self.data_table = None
        self.data_table_dict = DotDict()
        self.test_case_data = TestCaseData()

    def _start_suite(self, suite: TestSuite, result):
        """Called when a test suite starts.
        Data and result are model objects representing the executed test suite and its execution results, respectively.

        :param suite: class robot.running.model.TestSuite(name='', doc='', metadata=None, source=None)
        :param result: NOT USED
        """
        try:
            self.suite_name = suite.longname
            self.template_test = suite.tests[0]
            self._update_config()
            self.suite_source = suite.source
            self._create_data_table()
            debug("[ DataDriver ] data Table created")
            self.template_keyword = self._get_template_keyword(suite)
            self._clean_template_test()
            test_list = self._get_filtered_test_list()
            if self._handle_pabot(test_list):
                suite.tests = []
                suite.keywords.setup = None
                suite.keywords.teardown = None
            else:
                suite.tests = test_list
            self._set_date_table_to_robot_variable()
            debug(f"[ DataDriver ] {len(test_list)} tests added.")
        except Exception as e:
            error(f'[ DataDriver ] Error in robot file:\n  File "{suite.source}", line 0')
            if self.reader_config.file:
                error(
                    f'In source file:\n'
                    f'  File "{self.reader_config.file}", line {e.row if hasattr(e, "row") else "0"}'  # type: ignore
                )
            debug(traceback.format_exc())
            raise e

    def _start_test(self, test: TestCase, result):
        BuiltIn().set_test_variable(
            '${DataDriver_TEST_DATA}',
            self.data_table_dict.get(test.name, {"ERROR": "Test Case not found..."}),
        )

    def _set_date_table_to_robot_variable(self):
        BuiltIn().set_suite_variable('${DataDriver_DATA_LIST}', self.data_table)
        self.data_table_dict = DotDict([(item.test_case_name, item) for item in self.data_table])
        BuiltIn().set_suite_variable('${DataDriver_DATA_DICT}', self.data_table_dict)

    def _get_all_tags(self):
        all_tags = set()
        for test_data in self.data_table:
            all_tags.update(test_data.tags or [])
        return all_tags

    def _clean_template_test(self):
        if self.handle_template_tags == TagHandling.NoTags:
            self.template_test.tags = Tags()
        elif self.handle_template_tags == TagHandling.UnsetTags:
            for tag in self._get_all_tags():
                self.template_test.tags.remove(tag)
        if self._is_new_model():
            self.template_test.body = None
        else:
            setup = self.template_test.keywords.setup
            teardown = self.template_test.keywords.teardown
            self.template_test.keywords.clear()
            self.template_test.keywords.setup = setup
            self.template_test.keywords.teardown = teardown

    def _update_config(self):
        if self.config_dict.config_keyword:
            config = self.config_dict
            config_update = BuiltIn().run_keyword(self.config_dict.config_keyword, config)
            self.reader_config = ReaderConfig(**{**config, **config_update})

    def _get_filtered_test_list(self):
        temp_test_list = list()
        temp_data_table = list()
        dynamic_test_list = get_filter_dynamic_test_names()
        for index, self.test_case_data in enumerate(self.data_table):
            if self._included_by_tags() and self._not_excluded_by_tags():
                self._create_test_from_template()
                if (
                    dynamic_test_list is None
                    or f"{self.test.parent.name}.{self.test.name}" in dynamic_test_list
                    or self.test.longname in dynamic_test_list
                ):
                    temp_test_list.append(self.test)
                    temp_data_table.append(self.test_case_data)
        self.data_table = temp_data_table
        return temp_test_list

    def _included_by_tags(self):
        if self.include and isinstance(self.test_case_data.tags, list):
            return self._filter_tag(self.include)
        return True

    def _not_excluded_by_tags(self):
        if self.exclude and isinstance(self.test_case_data.tags, list):
            return not self._filter_tag(self.exclude)
        return True

    def _filter_tag(self, filter_setting):
        tags = Tags()
        if (
            self.handle_template_tags != TagHandling.DefaultTags
            or len(self.test_case_data.tags) == 0
        ):
            tags.add(self.template_test.tags)
        tags.add(self.test_case_data.tags)
        return tags.match(filter_setting)

    def _create_data_table(self):
        """
        this function creates a dictionary which contains all data from data file.
        Keys are header names.
        Values are data of this column as array.
        """
        self._resolve_file_attribute()
        self.data_table = self._data_reader().get_data_from_source()
        debug(f"[ DataDriver ] Opening file '{self.reader_config.file}'")
        debug(f"[ DataDriver ] {len(self.data_table)} Test Cases loaded...")

    def _data_reader(self) -> AbstractReaderClass:
        if not self.reader_config.reader_class:
            reader_class = self._get_data_reader_from_file_extension()
        else:
            reader_class = self._get_data_reader_from_reader_class()
        reader_instance = reader_class(self.reader_config)
        if not isinstance(reader_instance, AbstractReaderClass):
            raise ImportError(
                f"{self.reader_config.reader_class} in no instance of AbstractDataReader!"
            )
        return reader_instance

    def _get_data_reader_from_file_extension(self):
        file_extension = os.path.splitext(self.reader_config.file)[1]
        reader_type = file_extension.lower()[1:]
        debug(f"[ DataDriver ] Initialized in {reader_type}-mode.")
        reader_module = importlib.import_module(f"..{reader_type}_reader", "DataDriver.DataDriver")
        debug(f"[ DataDriver ] Reader Module: {reader_module}")
        reader_class = getattr(reader_module, f"{reader_type}_reader")
        return reader_class

    def _get_data_reader_from_reader_class(self):
        reader_name = self.reader_config.reader_class
        debug(f"[ DataDriver ] Initializes  {reader_name}")
        if os.path.isfile(reader_name):
            reader_class = self._get_reader_class_from_path(reader_name)
        else:
            local_file = os.path.join(os.path.split(os.path.realpath(__file__))[0], reader_name)
            relative_file = os.path.join(
                os.path.realpath(os.path.split(self.suite_source)[0]), reader_name
            )
            if os.path.isfile(local_file):
                reader_class = self._get_reader_class_from_path(local_file)
            elif os.path.isfile(relative_file):
                reader_class = self._get_reader_class_from_path(relative_file)
            else:
                try:
                    reader_class = self._get_reader_class_from_module(reader_name)
                except Exception as e:
                    reader_module = importlib.import_module(
                        f"..{reader_name}", "DataDriver.DataDriver"
                    )
                    reader_class = getattr(reader_module, reader_name)
        debug(f"[ DataDriver ] Reader Class: {reader_class}")
        return reader_class

    @staticmethod
    def _get_reader_class_from_path(file_name):
        debug(f"[ DataDriver ] Loading Reader from file {file_name}")
        abs_path = os.path.abspath(file_name)
        importer = Importer("DataReader")
        debug(f"[ DataDriver ] Reader path: {abs_path}")
        reader = importer.import_class_or_module_by_path(abs_path)
        if not inspect.isclass(reader):
            message = f"Importing custom DataReader class from {abs_path} failed."
            raise ImportError(message)
        return reader

    @staticmethod
    def _get_reader_class_from_module(reader_name):
        importer = Importer("DataReader")
        debug(f"[ DataDriver ] Reader Module: {reader_name}")
        reader = importer.import_class_or_module(reader_name)
        if not inspect.isclass(reader):
            message = f"Importing custom DataReader class {reader_name} failed."
            raise ImportError(message)
        return reader

    def _resolve_file_attribute(self) -> None:
        if self.reader_config.file_search_strategy == "PATH":
            if self.reader_config.reader_class and not self.reader_config.file:
                return
            if self._check_valid_glob():
                return
            if (not self.reader_config.file) or (
                "" == self.reader_config.file[: self.reader_config.file.rfind(".")]
            ):
                self._set_data_file_to_suite_source()
            else:
                self._check_if_file_exists_as_path_or_in_suite()
        elif self.reader_config.file_search_strategy == "REGEX":
            self._search_file_from_regex()
        elif self.reader_config.file_search_strategy == "NONE":
            pass  # If file_search_strategy is None, no validation of the input file is done. Use i.e. for SQL sources.
        else:
            raise ValueError(
                f"file_search_strategy={self.reader_config.file_search_strategy} is not a valid value!"
            )

    def _set_data_file_to_suite_source(self):
        if not self.reader_config.file:
            suite_path_as_data_file = f"{self.suite_source[:self.suite_source.rfind('.')]}.csv"
        else:
            suite_path = self.suite_source[: self.suite_source.rfind(".")]
            file_extension = self.reader_config.file[self.reader_config.file.rfind(".") :]
            suite_path_as_data_file = f"{suite_path}{file_extension}"
        if os.path.isfile(suite_path_as_data_file):
            self.reader_config.file = suite_path_as_data_file
        else:
            raise FileNotFoundError(
                f"File attribute was empty. "
                f"Tried to find {suite_path_as_data_file} but file does not exist. "
                f"If no file validation is required, set file_search_strategy=None."
            )

    def _check_if_file_exists_as_path_or_in_suite(self):
        if not os.path.isfile(self.reader_config.file):
            suite_dir = str(os.path.dirname(self.suite_source))
            file_in_suite_dir = os.path.join(suite_dir, self.reader_config.file)
            if os.path.isfile(file_in_suite_dir):
                self.reader_config.file = file_in_suite_dir
            else:
                raise FileNotFoundError(
                    f"File attribute was not a full path. Tried to find {file_in_suite_dir} but file does not exist."
                )

    def _check_valid_glob(self):
        if not self.reader_config.reader_class == "glob_reader":
            return
        if not glob(self.reader_config.file):
            raise FileNotFoundError(
                f"Glob pattern did not find a file or folder. Glob pattern was: {self.reader_config.file}"
            )
        return True

    def _search_file_from_regex(self):
        if os.path.isdir(self.reader_config.file):
            for filename in os.listdir(self.reader_config.file):
                if re.match(self.reader_config.file_regex, filename):
                    self.reader_config.file = os.path.join(self.reader_config.file, filename)
                    break

    def _handle_pabot(self, test_list):
        if (
            get_variable_value("${DYNAMICTEST}")
            or get_variable_value("${DYNAMICTESTS}")
            or not self.robot_options["test"]
            or not get_variable_value("${PABOTQUEUEINDEX}")
        ):
            return
        pabot_process_count = int(get_variable_value("${PABOTNUMBEROFPROCESSES}"))
        if not pabot_process_count:
            warn(
                "You are using an incompatible version of Pabot! "
                " '--testlevelsplit' is not supported between Pabot 1.2.1 and 1.10.1 with DataDriver"
            )
            return
        try:
            from pabot.pabotlib import Remote  # type: ignore
        except ImportError as e:
            debug(e)
            return
        pabotlib_url = get_variable_value("${PABOTLIBURI}")
        try:
            pabotlib = Remote(pabotlib_url)
            if not pabotlib:
                raise ConnectionError
            self._create_pabot_queue(pabot_process_count, pabotlib, test_list)
        except (RuntimeError, ConnectionError) as e:
            error(e)
            error(
                f"Unable to connect to PabotLib via '{pabotlib_url}'! "
                f"Is PabotLib in use? Try 'pabot --pabotlib'"
            )
            error("Execution as been processes without --testlevelsplit")
            return
        pabotlib.run_keyword("ignore_execution", [get_variable_value("${CALLER_ID}")], {})
        return True

    def _create_pabot_queue(self, pabot_process_count, pabotlib, test_list):
        pabot_opt = self.reader_config.optimize_pabot
        if pabot_opt == PabotOpt.Atomic:
            self._add_test_to_pabot_queue(pabotlib, test_list)
        else:
            if pabot_opt == PabotOpt.Binary:
                partitioner = binary_partition_test_list
            else:
                partitioner = equally_partition_test_list
            for process_test_list in partitioner(test_list, pabot_process_count):
                self._add_test_list_to_pabot_queue(pabotlib, process_test_list)

    def _add_test_list_to_pabot_queue(self, pabotlib, test_list):
        test_names = [f"{self.suite_name}.{test.name}" for test in test_list]
        pabotlib.run_keyword(
            "add_suite_to_execution_queue",
            [self.suite_name, [f"DYNAMICTESTS:{'|'.join(test_names)}"]],
            {},
        )

    def _add_test_to_pabot_queue(self, pabotlib, test_list):
        for test in test_list:
            pabotlib.run_keyword(
                "add_suite_to_execution_queue",
                [self.suite_name, [f"DYNAMICTEST:{self.suite_name}.{test.name}"]],
                {},
            )

    def _get_template_keyword(self, suite):
        template = self.template_test.template
        if template:
            for keyword in suite.resource.keywords:
                if is_same_keyword(keyword.name, template):
                    return keyword
        raise AttributeError('No "Test Template" keyword found for first test case.')

    def _create_test_from_template(self):
        self.test = TestCase(
            name=self.template_test.name,
            doc=self.template_test.doc,
            tags=self.template_test.tags,
            template=self.template_test.template,
            lineno=self.template_test.lineno,
        )
        self.test.parent = self.template_test.parent
        self._replace_test_case_name()
        self._replace_test_case_keywords()
        self._add_test_case_tags()
        self._replace_test_case_doc()

    def _replace_test_case_name(self):
        if self.test_case_data.test_case_name == "":
            for variable_name in self.test_case_data.arguments:
                self.test.name = self.test.name.replace(
                    variable_name, str(self.test_case_data.arguments[variable_name])
                )
            self.test_case_data.test_case_name = self.test.name
        else:
            self.test.name = self.test_case_data.test_case_name

    def _replace_test_case_keywords(self):
        if self._is_new_model():
            self.test.setup = self.template_test.setup
            self.test.teardown = self.template_test.teardown
            self.test.body.create_keyword(
                name=self.template_keyword.name, args=self._get_template_arguments()
            )
        else:
            self.test.keywords.setup = self.template_test.keywords.setup
            self.test.keywords.create(
                name=self.template_keyword.name, args=self._get_template_arguments()
            )
            self.test.keywords.teardown = self.template_test.keywords.teardown

    def _get_template_arguments(self):
        return_arguments = []
        for arg in self.template_keyword.args:
            if arg in self.test_case_data.arguments:
                return_arguments.append(self.test_case_data.arguments[arg])
            else:
                return_arguments.append(arg)
        return return_arguments

    def _add_test_case_tags(self):
        if isinstance(self.test_case_data.tags, list):
            if (
                self.handle_template_tags == TagHandling.DefaultTags
                and len(self.test_case_data.tags) > 0
            ):
                self.test.tags = Tags()
            for tag in self.test_case_data.tags:
                self.test.tags.add(tag.strip())
        self._add_tag_if_pabot_dryrun()

    def _add_tag_if_pabot_dryrun(self):
        if is_pabot_dry_run():
            self.test.tags.add("pabot:dynamictest")

    def _replace_test_case_doc(self):
        self.test.doc = self.test_case_data.documentation

    def _is_new_model(self):
        return hasattr(self.template_test, "body")
